import json
import typing as t

import pytest
from jupyter_server.serverapp import ServerApp
from tornado.httpclient import HTTPResponse

from jupyterlab_code_formatter import load_jupyter_server_extension


@pytest.fixture(autouse=True)
def jcf_serverapp(jp_serverapp: ServerApp) -> ServerApp:
    load_jupyter_server_extension(jp_serverapp)
    return jp_serverapp


@pytest.fixture
def request_format(jp_fetch):  # type: ignore[no-untyped-def]
    def do_request(
        formatter: str,
        code: t.List[str],
        options: t.Dict[str, t.Any],
        headers: t.Optional[t.Dict[str, t.Any]] = None,
        **kwargs: t.Any,
    ) -> HTTPResponse:
        return jp_fetch(  # type: ignore[no-any-return]
            "jupyterlab_code_formatter",
            "format",
            method="POST",
            body=json.dumps(
                {
                    "code": code,
                    "options": options,
                    "notebook": True,
                    "formatter": formatter,
                }
            ),
            headers=headers,
            **kwargs,
        )

    return do_request


@pytest.fixture
def request_list_formatters(jp_fetch):  # type: ignore[no-untyped-def]
    def do_request(
        headers: t.Optional[t.Dict[str, t.Any]] = None,
        **kwargs: t.Any,
    ) -> HTTPResponse:
        return jp_fetch(  # type: ignore[no-any-return]
            "jupyterlab_code_formatter",
            "formatters",
            method="GET",
            headers=headers,
            **kwargs,
        )

    return do_request



import json

import tornado
from jupyter_server.base.handlers import APIHandler
from jupyter_server.utils import url_path_join

from jupyterlab_code_formatter.formatters import SERVER_FORMATTERS


class FormattersAPIHandler(APIHandler):
    @tornado.web.authenticated
    def get(self) -> None:
        """Show what formatters are installed and available."""
        use_cache = self.get_query_argument("cached", default=None)
        self.finish(
            json.dumps(
                {
                    "formatters": {
                        name: {
                            "enabled": formatter.cached_importable if use_cache else formatter.importable,
                            "label": formatter.label,
                        }
                        for name, formatter in SERVER_FORMATTERS.items()
                    }
                }
            )
        )


class FormatAPIHandler(APIHandler):
    @tornado.web.authenticated
    def post(self) -> None:
        data = json.loads(self.request.body.decode("utf-8"))
        formatter_instance = SERVER_FORMATTERS.get(data["formatter"])
        use_cache = self.get_query_argument("cached", default=None)

        if formatter_instance is None or not (
            formatter_instance.cached_importable if use_cache else formatter_instance.importable
        ):
            self.set_status(404, f"Formatter {data['formatter']} not found!")
            self.finish()
        else:
            notebook = data["notebook"]
            options = data.get("options", {})
            formatted_code = []
            for code in data["code"]:
                try:
                    formatted_code.append({"code": formatter_instance.format_code(code, notebook, **options)})
                except Exception as e:
                    formatted_code.append({"error": str(e)})
            self.finish(json.dumps({"code": formatted_code}))


def setup_handlers(web_app):
    host_pattern = ".*$"

    base_url = web_app.settings["base_url"]

    web_app.add_handlers(
        host_pattern,
        [
            (
                url_path_join(base_url, "jupyterlab_code_formatter/formatters"),
                FormattersAPIHandler,
            )
        ],
    )

    web_app.add_handlers(
        host_pattern,
        [
            (
                url_path_join(base_url, "/jupyterlab_code_formatter/format"),
                FormatAPIHandler,
            )
        ],
    )



import json
import sys
import typing as t
if sys.version_info >= (3, 8):
    from importlib.metadata import version
else:
    from importlib_metadata import version

import pytest
from jsonschema import validate
from tornado.httpclient import HTTPResponse

from jupyterlab_code_formatter.formatters import SERVER_FORMATTERS


def _generate_list_formaters_entry_json_schema(
    formatter_name: str,
) -> t.Dict[str, t.Any]:
    return {
        "type": "object",
        "required": [formatter_name],
        "properties": {
            formatter_name: {
                "type": "object",
                "required": ["enabled", "label"],
                "properties": {
                    "enabled": {"type": "boolean"},
                    "label": {"type": "string"},
                },
            }
        },
    }


EXPECTED_LIST_FORMATTERS_SCHEMA = {
    "type": "object",
    "required": ["formatters"],
    "properties": {
        "formatters": {
            formatter_name: _generate_list_formaters_entry_json_schema(formatter_name)
            for formatter_name in SERVER_FORMATTERS
        }
    },
}
EXPECTED_FROMAT_SCHEMA = {
    "type": "object",
    "required": ["code"],
    "properties": {
        "code": {
            "type": "array",
            "items": {
                "type": "object",
                "oneOf": [
                    {
                        "additionalProperties": False,
                        "properties": {"code": {"type": "string"}},
                    },
                    {
                        "additionalProperties": False,
                        "properties": {"error": {"type": "string"}},
                    },
                ],
            },
        }
    },
}


SIMPLE_VALID_PYTHON_CODE = "x= 22;  e          =1"


def _check_http_code_and_schema(
    response: HTTPResponse, expected_code: int, expected_schema: t.Dict[str, t.Any]
) -> t.Dict[str, t.Any]:
    assert response.code == expected_code
    json_result: t.Dict[str, t.Any] = json.loads(response.body)
    validate(instance=json_result, schema=expected_schema)
    return json_result


async def test_list_formatters(request_list_formatters):  # type: ignore[no-untyped-def]
    """Check if the formatters list route works."""
    response: HTTPResponse = await request_list_formatters()
    _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_LIST_FORMATTERS_SCHEMA,
    )


async def test_404_on_unknown(request_format):  # type: ignore[no-untyped-def]
    """Check that it 404 correctly if formatter name is bad."""
    response: HTTPResponse = await request_format(
        formatter="UNKNOWN",
        code=[SIMPLE_VALID_PYTHON_CODE],
        options={},
        raise_error=False,
    )
    assert response.code == 404


async def test_can_apply_python_formatter(request_format):  # type: ignore[no-untyped-def]
    """Check that it can apply black with simple config."""
    response: HTTPResponse = await request_format(
        formatter="black",
        code=[SIMPLE_VALID_PYTHON_CODE],
        options={"line_length": 88},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == "x = 22\ne = 1"


async def test_can_use_black_config(request_format):  # type: ignore[no-untyped-def]
    """Check that it can apply black with advanced config."""
    given = "some_string='abc'"
    expected = "some_string = 'abc'"

    response: HTTPResponse = await request_format(
        formatter="black",
        code=[given],
        options={"line_length": 123, "string_normalization": False},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


async def test_return_error_if_any(request_format):  # type: ignore[no-untyped-def]
    """Check that it returns the error if any."""
    bad_python = "this_is_bad = 'hihi"

    response: HTTPResponse = await request_format(
        formatter="black",
        code=[bad_python],
        options={"line_length": 123, "string_normalization": False},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["error"] == "Cannot parse: 1:13: this_is_bad = 'hihi"


@pytest.mark.parametrize("formatter", ("black", "yapf", "isort"))
async def test_can_handle_magic(request_format, formatter):  # type: ignore[no-untyped-def]
    """Check that it's fine to run formatters for code with magic."""
    given = '%%timeit\nsome_string = "abc"'
    expected = '%%timeit\nsome_string = "abc"'

    response: HTTPResponse = await request_format(
        formatter=formatter,
        code=[given],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


@pytest.mark.parametrize("formatter", ("black", "yapf", "isort"))
async def test_can_handle_shell_cmd(request_format, formatter):  # type: ignore[no-untyped-def]
    """Check that it's fine to run formatters for code with shell cmd."""
    given = '%%timeit\nsome_string = "abc"\n!pwd'
    expected = '%%timeit\nsome_string = "abc"\n!pwd'

    response: HTTPResponse = await request_format(
        formatter=formatter,
        code=[given],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


@pytest.mark.parametrize("formatter", ("black", "yapf", "isort"))
async def test_can_handle_incompatible_magic_language(request_format, formatter):  # type: ignore[no-untyped-def]
    """Check that it will ignore incompatible magic language cellblock."""
    given = "%%html\n<h1>Hi</h1>"
    expected = "%%html\n<h1>Hi</h1>"

    response: HTTPResponse = await request_format(
        formatter=formatter,
        code=[given],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


@pytest.mark.parametrize("formatter", ("black", "yapf", "isort"))
async def test_can_handle_incompatible_magic_language_single(request_format, formatter):  # type: ignore[no-untyped-def]
    """Check that it will ignore incompatible magic language cellblock with single %."""
    given = "%html <h1>Hi</h1>"
    expected = "%html <h1>Hi</h1>"

    response: HTTPResponse = await request_format(
        formatter=formatter,
        code=[given],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


async def test_can_ipython_help_signle(request_format):  # type: ignore[no-untyped-def]
    """Check that it will ignore single question mark interactive help lines on the fly."""
    given = "    bruh?\nprint('test')\n#test?"
    expected = '    bruh?\nprint("test")\n# test?'

    response: HTTPResponse = await request_format(
        formatter="black",
        code=[given],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


async def test_can_ipython_help_double(request_format):  # type: ignore[no-untyped-def]
    """Check that it will ignore double question mark interactive help lines on the fly."""
    given = "    bruh??\nprint('test')\n#test?"
    expected = '    bruh??\nprint("test")\n# test?'

    response: HTTPResponse = await request_format(
        formatter="black",
        code=[given],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


async def test_can_ipython_help_signle_leading(request_format):  # type: ignore[no-untyped-def]
    """Check that it will ignore leading single question mark interactive help lines on the fly."""
    given = "    ?bruh\nprint('test')\n#test?"
    expected = '    ?bruh\nprint("test")\n# test?'

    response: HTTPResponse = await request_format(
        formatter="black",
        code=[given],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


async def test_can_ipython_help_double_leading(request_format):  # type: ignore[no-untyped-def]
    """Check that it will ignore leading double question mark interactive help lines on the fly."""
    given = "    ??bruh\nprint('test')\n#test?"
    expected = '    ??bruh\nprint("test")\n# test?'

    response: HTTPResponse = await request_format(
        formatter="black",
        code=[given],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


async def test_will_ignore_quarto_comments(request_format):  # type: ignore[no-untyped-def]
    """Check that it will ignore Quarto's comments at the top of a block."""
    given = """#| eval: false
1 + 1"""

    response: HTTPResponse = await request_format(
        formatter="black",
        code=[given],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == given


async def test_will_ignore_run_command(request_format):  # type: ignore[no-untyped-def]
    """Check that it will ignore run command."""
    given = "     run     some_script.py"

    response: HTTPResponse = await request_format(
        formatter="black",
        code=[given],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == given


async def test_will_ignore_question_mark(request_format):  # type: ignore[no-untyped-def]
    """Check that it will ignore single question mark in comments."""
    given = """def f():
    # bruh what?
    # again bruh? really
    # a ? b
    print('hi')
    x = '?'"""
    expected = """def f():
    # bruh what?
    # again bruh? really
    # a ? b
    print("hi")
    x = "?\""""

    response: HTTPResponse = await request_format(
        formatter="black",
        code=[given],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


async def test_will_ignore_question_mark2(request_format):  # type: ignore[no-untyped-def]
    """Check that it will ignore double question mark in comments."""
    given = """def f():
    # bruh what??
    # again bruh?? really
    # a ? b ? c
    print('hi')"""
    expected = """def f():
    # bruh what??
    # again bruh?? really
    # a ? b ? c
    print("hi")"""

    response: HTTPResponse = await request_format(
        formatter="black",
        code=[given],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


async def test_will_ignore_question_weird(request_format):  # type: ignore[no-untyped-def]
    given = """wat
wat??"""
    expected = """wat
wat??"""

    response: HTTPResponse = await request_format(
        formatter="black",
        code=[given],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


async def test_can_use_styler(request_format):  # type: ignore[no-untyped-def]
    given = "a = 3; 2"
    expected = "a <- 3\n2"

    response: HTTPResponse = await request_format(
        formatter="styler",
        code=[given],
        options={"scope": "tokens"},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


async def test_can_use_styler2(request_format):  # type: ignore[no-untyped-def]
    given = """data_frame(
     small  = 2 ,
     medium = 4,#comment without space
     large  =6
)"""
    expected = """data_frame(
  small  = 2,
  medium = 4, # comment without space
  large  = 6
)"""

    response: HTTPResponse = await request_format(
        formatter="styler",
        code=[given],
        options={"strict": False},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


async def test_can_use_styler3(request_format):  # type: ignore[no-untyped-def]
    given = "1++1/2*2^2"
    expected = "1 + +1/2*2^2"

    response: HTTPResponse = await request_format(
        formatter="styler",
        code=[given],
        options={
            "math_token_spacing": {
                "one": ["'+'", "'-'"],
                "zero": ["'/'", "'*'", "'^'"],
            }
        },
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


async def test_can_use_styler4(request_format):  # type: ignore[no-untyped-def]
    given = """a <- function() {
    ### not to be indented
    # indent normally
    33
    }"""
    expected = """a <- function() {
### not to be indented
  # indent normally
  33
}"""

    response: HTTPResponse = await request_format(
        formatter="styler",
        code=[given],
        options={
            "reindention": {
                "regex_pattern": "^###",
                "indention": 0,
                "comments_only": True,
            }
        },
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


async def test_can_use_styler5(request_format):  # type: ignore[no-untyped-def]
    given = """call(
#          SHOULD BE ONE SPACE BEFORE
1,2)
"""
    expected = """call(
    # SHOULD BE ONE SPACE BEFORE
    1, 2
)"""

    response: HTTPResponse = await request_format(
        formatter="styler",
        code=[given],
        options={"indent_by": 4, "start_comments_with_one_space": True},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


async def test_can_use_styler6(request_format):  # type: ignore[no-untyped-def]
    given = "1+1-3"
    expected = "1 + 1 - 3"

    response: HTTPResponse = await request_format(
        formatter="styler",
        code=[given],
        options={
            "math_token_spacing": "tidyverse_math_token_spacing",
            "reindention": "tidyverse_reindention",
        },
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


@pytest.mark.skip(reason="rust toolchain doesn't seem to be picked up here for some reason.")
async def test_can_rustfmt(request_format):  # type: ignore[no-untyped-def]
    given = """// function to add two numbers
fn add() {
    let a = 5;
                let b = 10;

            let sum = a + b;

    println!("Sum of a and b = {}", 
        sum);
}

fn main() {
    // function call
    add();
}"""
    expected = """// function to add two numbers
fn add() {
    let a = 5;
    let b = 10;

    let sum = a + b;

    println!("Sum of a and b = {}", sum);
}

fn main() {
    // function call
    add();
}"""

    response: HTTPResponse = await request_format(
        formatter="rustfmt",
        code=[given],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == expected


@pytest.mark.xfail(reason="Rust toolchain isn't respected in test for some reason atm.")
async def test_can_apply_ruff(request_format):  # type: ignore[no-untyped-def]
    """Check that it can apply black with simple config."""
    response: HTTPResponse = await request_format(
        formatter="ruff",
        code=[SIMPLE_VALID_PYTHON_CODE],
        options={},
    )
    json_result = _check_http_code_and_schema(
        response=response,
        expected_code=200,
        expected_schema=EXPECTED_FROMAT_SCHEMA,
    )
    assert json_result["code"][0]["code"] == "x = 22\ne = 1"



from ._version import __version__
from .handlers import setup_handlers


def _jupyter_labextension_paths():
    return [{"src": "labextension", "dest": "jupyterlab_code_formatter"}]


def _jupyter_server_extension_points():
    return [{"module": "jupyterlab_code_formatter"}]


def _load_jupyter_server_extension(server_app):
    """Registers the API handler to receive HTTP requests from the frontend extension.

    Parameters
    ----------
    server_app: jupyterlab.labapp.LabApp
        JupyterLab application instance
    """
    setup_handlers(server_app.web_app)
    name = "jupyterlab_code_formatter"
    server_app.log.info(f"Registered {name} server extension")


# For backward compatibility with notebook server - useful for Binder/JupyterHub
load_jupyter_server_extension = _load_jupyter_server_extension



import abc
import copy
import importlib
import logging
import re
import shutil
import subprocess
import sys
from functools import wraps
from typing import List, Type

try:
    import rpy2
    import rpy2.robjects
except ImportError:
    pass
if sys.version_info >= (3, 9):
    from functools import cache
else:
    from functools import lru_cache
    cache = lru_cache(maxsize=None)

from packaging import version

logger = logging.getLogger(__name__)


INCOMPATIBLE_MAGIC_LANGUAGES = [
    "html",
    "js",
    "javascript",
    "latex",
    "perl",
    "markdown",
    "ruby",
    "script",
    "sh",
    "svg",
    "bash",
    "info",
    "cleanup",
    "delete",
    "configure",
    "logs",
    "sql",
    "local",
    "sparksql",
]


class BaseFormatter(abc.ABC):
    @property
    @abc.abstractmethod
    def label(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def importable(self) -> bool:
        pass

    @abc.abstractmethod
    def format_code(self, code: str, notebook: bool, **options) -> str:
        pass

    @property
    @cache
    def cached_importable(self) -> bool:
        return self.importable


class BaseLineEscaper(abc.ABC):
    """A base class for defining how to escape certain sequence of text to avoid formatting."""

    def __init__(self, code: str) -> None:
        self.code = code

    @property
    @abc.abstractmethod
    def langs(self) -> List[str]:
        pass

    @abc.abstractmethod
    def escape(self, line: str) -> str:
        pass

    @abc.abstractmethod
    def unescape(self, line: str) -> str:
        pass


class MagicCommandEscaper(BaseLineEscaper):
    langs = ["python"]
    escaped_line_start = "# \x01 "
    unesacpe_start = len(escaped_line_start)

    def escape(self, line: str) -> str:
        if line.lstrip().startswith("%"):
            line = f"{self.escaped_line_start}{line}"
        return line

    def unescape(self, line: str) -> str:
        if line.lstrip().startswith(self.escaped_line_start):
            line = line[self.unesacpe_start :]
        return line


class RunScriptEscaper(BaseLineEscaper):
    langs = ["python"]
    escaped_line_start = "# \x01 "
    unesacpe_start = len(escaped_line_start)

    def escape(self, line: str) -> str:
        if re.match(pattern="run\s+\w+", string=line.lstrip()):
            line = f"{self.escaped_line_start}{line}"
        return line

    def unescape(self, line: str) -> str:
        if line.lstrip().startswith(self.escaped_line_start):
            line = line[self.unesacpe_start :]
        return line


class HelpEscaper(BaseLineEscaper):
    langs = ["python"]
    escaped_line_start = "# \x01 "
    unesacpe_start = len(escaped_line_start)

    def escape(self, line: str) -> str:
        lstripped = line.lstrip()
        if (
            line.endswith("??")
            or line.endswith("?")
            or lstripped.startswith("?")
            or lstripped.startswith("??")
        ) and "#" not in line:
            line = f"{self.escaped_line_start}{line}"
        return line

    def unescape(self, line: str) -> str:
        if line.lstrip().startswith(self.escaped_line_start):
            line = line[self.unesacpe_start :]
        return line


class CommandEscaper(BaseLineEscaper):
    langs = ["python"]
    escaped_line_start = "# \x01 "
    unesacpe_start = len(escaped_line_start)

    def escape(self, line: str) -> str:
        if line.lstrip().startswith("!"):
            line = f"{self.escaped_line_start}{line}"
        return line

    def unescape(self, line: str) -> str:
        if line.lstrip().startswith(self.escaped_line_start):
            line = line[self.unesacpe_start :]
        return line


class QuartoCommentEscaper(BaseLineEscaper):
    langs = ["python"]
    escaped_line_start = "# \x01 "
    unesacpe_start = len(escaped_line_start)

    def escape(self, line: str) -> str:
        if line.lstrip().startswith("#| "):
            line = f"{self.escaped_line_start}{line}"
        return line

    def unescape(self, line: str) -> str:
        if line.lstrip().startswith(self.escaped_line_start):
            line = line[self.unesacpe_start :]
        return line


ESCAPER_CLASSES: List[Type[BaseLineEscaper]] = [
    MagicCommandEscaper,
    HelpEscaper,
    CommandEscaper,
    QuartoCommentEscaper,
    RunScriptEscaper,
]


def handle_line_ending_and_magic(func):
    @wraps(func)
    def wrapped(self, code: str, notebook: bool, **options) -> str:
        if any(code.startswith(f"%{lang}") for lang in INCOMPATIBLE_MAGIC_LANGUAGES) or any(
            code.startswith(f"%%{lang}") for lang in INCOMPATIBLE_MAGIC_LANGUAGES
        ):
            logger.info("Non compatible magic language cell block detected, ignoring.")
            return code

        has_semicolon = code.strip().endswith(";")

        escapers = [escaper_cls(code) for escaper_cls in ESCAPER_CLASSES]

        lines = code.splitlines()
        for escaper in escapers:
            lines = map(escaper.escape, lines)
        code = "\n".join(lines)

        code = func(self, code, notebook, **options)

        lines = code.splitlines()
        lines.append("")

        for escaper in escapers:
            lines = map(escaper.unescape, lines)
        code = "\n".join(lines)

        if notebook:
            code = code.rstrip()

        if has_semicolon and notebook and not code.endswith(";"):
            code += ";"
        return code

    return wrapped


BLUE_MONKEY_PATCHED = False


def is_importable(pkg_name: str) -> bool:
    # find_spec will check for packages installed/uninstalled after JupyterLab started
    return importlib.util.find_spec(pkg_name) is not None


def command_exist(name: str) -> bool:
    """Detect that command (executable) is installed"""
    return shutil.which(name) is not None


def import_black():
    global BLUE_MONKEY_PATCHED
    if BLUE_MONKEY_PATCHED:
        for module in list(sys.modules):
            if module.startswith("black."):
                importlib.reload(sys.modules[module])

        import black

        black = importlib.reload(black)
        BLUE_MONKEY_PATCHED = False
    else:
        import black

    return black


def import_blue():
    """Import blue and perform monkey patch."""
    global BLUE_MONKEY_PATCHED
    import blue

    if not BLUE_MONKEY_PATCHED:
        blue.monkey_patch_black(blue.Mode.synchronous)
        BLUE_MONKEY_PATCHED = True

    return blue


class BlueFormatter(BaseFormatter):
    label = "Apply Blue Formatter"

    @property
    def importable(self) -> bool:
        return is_importable("blue")

    @staticmethod
    def handle_options(**options):
        blue = import_blue()

        return {"mode": blue.black.FileMode(**options)}

    @handle_line_ending_and_magic
    def format_code(self, code: str, notebook: bool, **options) -> str:
        blue = import_blue()

        code = blue.black.format_str(code, **self.handle_options(**options))
        return code


class BlackFormatter(BaseFormatter):
    label = "Apply Black Formatter"

    @property
    def importable(self) -> bool:
        return is_importable("black")

    @staticmethod
    def handle_options(**options):
        black = import_black()

        file_mode_change_version = version.parse("19.3b0")
        current_black_version = version.parse(black.__version__)
        if current_black_version >= file_mode_change_version:
            return {"mode": black.FileMode(**options)}
        else:
            return options

    @handle_line_ending_and_magic
    def format_code(self, code: str, notebook: bool, **options) -> str:
        black = import_black()

        code = black.format_str(code, **self.handle_options(**options))
        return code


class Autopep8Formatter(BaseFormatter):
    label = "Apply Autopep8 Formatter"

    @property
    def importable(self) -> bool:
        return is_importable("autopep8")

    @handle_line_ending_and_magic
    def format_code(self, code: str, notebook: bool, **options) -> str:
        from autopep8 import fix_code

        return fix_code(code, options=options)


class YapfFormatter(BaseFormatter):
    label = "Apply YAPF Formatter"

    @property
    def importable(self) -> bool:
        return is_importable("yapf")

    @handle_line_ending_and_magic
    def format_code(self, code: str, notebook: bool, **options) -> str:
        from yapf.yapflib.yapf_api import FormatCode

        return FormatCode(code, **options)[0]


class IsortFormatter(BaseFormatter):
    label = "Apply Isort Formatter"

    @property
    def importable(self) -> bool:
        return is_importable("isort")

    @handle_line_ending_and_magic
    def format_code(self, code: str, notebook: bool, **options) -> str:
        try:
            from isort import SortImports

            return SortImports(file_contents=code, **options).output
        except ImportError:
            import isort

            return isort.code(code=code, **options)


class FormatRFormatter(BaseFormatter):
    label = "Apply FormatR Formatter"
    package_name = "formatR"

    @property
    def importable(self) -> bool:
        try:
            import rpy2.robjects.packages as rpackages

            rpackages.importr(self.package_name, robject_translations={".env": "env"})

            return True
        except Exception:
            return False

    @handle_line_ending_and_magic
    def format_code(self, code: str, notebook: bool, **options) -> str:
        import rpy2.robjects.packages as rpackages

        format_r = rpackages.importr(self.package_name, robject_translations={".env": "env"})
        formatted_code = format_r.tidy_source(text=code, output=False, **options)
        return "\n".join(formatted_code[0])


class StylerFormatter(BaseFormatter):
    label = "Apply Styler Formatter"
    package_name = "styler"

    @property
    def importable(self) -> bool:
        try:
            import rpy2.robjects.packages as rpackages

            rpackages.importr(self.package_name)

            return True
        except Exception:
            return False

    @handle_line_ending_and_magic
    def format_code(self, code: str, notebook: bool, **options) -> str:
        import rpy2.robjects.packages as rpackages

        styler_r = rpackages.importr(self.package_name)
        formatted_code = styler_r.style_text(code, **self._transform_options(styler_r, options))
        return "\n".join(formatted_code)

    @staticmethod
    def _transform_options(styler_r, options):
        transformed_options = copy.deepcopy(options)

        if "math_token_spacing" in transformed_options:
            if isinstance(options["math_token_spacing"], dict):
                transformed_options["math_token_spacing"] = rpy2.robjects.ListVector(
                    options["math_token_spacing"]
                )
            else:
                transformed_options["math_token_spacing"] = rpy2.robjects.ListVector(
                    getattr(styler_r, options["math_token_spacing"])()
                )

        if "reindention" in transformed_options:
            if isinstance(options["reindention"], dict):
                transformed_options["reindention"] = rpy2.robjects.ListVector(options["reindention"])
            else:
                transformed_options["reindention"] = rpy2.robjects.ListVector(
                    getattr(styler_r, options["reindention"])()
                )
        return transformed_options


class CommandLineFormatter(BaseFormatter):
    command: List[str]

    def __init__(self, command: List[str]):
        self.command = command

    @property
    def label(self) -> str:
        return f"Apply {self.command[0]} Formatter"

    @property
    def importable(self) -> bool:
        return command_exist(self.command[0])

    @handle_line_ending_and_magic
    def format_code(self, code: str, notebook: bool, args: List[str] = [], **options) -> str:
        process = subprocess.run(
            self.command + args,
            input=code,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        if process.stderr:
            logger.info(f"An error with {self.command[0]} has occurred:")
            logger.info(process.stderr)
            return code
        else:
            return process.stdout


class RuffFixFormatter(CommandLineFormatter):
    @property
    def label(self) -> str:
        return f"Apply ruff Formatter"

    def __init__(self):
        try:
            from ruff.__main__ import find_ruff_bin

            ruff_command = find_ruff_bin()
        except (ImportError, FileNotFoundError):
            ruff_command = "ruff"
        self.command = [ruff_command, "check", "--fix-only", "-"]


SERVER_FORMATTERS = {
    "black": BlackFormatter(),
    "blue": BlueFormatter(),
    "autopep8": Autopep8Formatter(),
    "yapf": YapfFormatter(),
    "isort": IsortFormatter(),
    "ruff": RuffFixFormatter(),
    "formatR": FormatRFormatter(),
    "styler": StylerFormatter(),
    "scalafmt": CommandLineFormatter(command=["scalafmt", "--stdin"]),
    "rustfmt": CommandLineFormatter(command=["rustfmt"]),
    "astyle": CommandLineFormatter(command=["astyle"]),
}
